/*
 * ALSA LAMIP-0.0.3 output plugin for kernel 2.6.x
 *
*/

#define ALSA_PCM_NEW_HW_PARAMS_API
#define ALSA_PCM_NEW_SW_PARAMS_API
#include <alsa/asoundlib.h>
#include <plug_out.h>

static int	alsa_init(const lamipPluginHandle*);
static void	alsa_cleanup(void);
static int	alsa_open(const lamipPCMInfo*);
static void	alsa_close(void);
static size_t	alsa_write_audio(const void*, size_t);
static void	alsa_drain(void);
static void	alsa_drop(void);


#ifdef DEBUGING
#define DEBUG(a) fprintf(stderr, "DEBUG_ALSA: %s\n", a)
#else
#define DEBUG(a)
#endif


static char*			device		= 0;
static snd_pcm_t*		handle		= 0;
static snd_pcm_format_t		alsa_format	= 0;
static snd_pcm_hw_params_t*	hwparams	= 0;
static char* 			card_id		= 0;
static int			nb_channel_tmp	= 0;

static OutputPlugin     alsa_functions;

/*****************************************************************************/

OutputPlugin*	lamip_output_info(void)
{
  alsa_functions.common.name		= "alsa";
  alsa_functions.common.description	= "ALSA output - Frederic Fondriest for LAMIP-0.0.3";
  alsa_functions.common.init		= alsa_init;
  alsa_functions.common.cleanup		= alsa_cleanup;
  alsa_functions.open			= alsa_open;
  alsa_functions.close			= alsa_close;
  alsa_functions.write			= alsa_write_audio;
  alsa_functions.drop			= alsa_drop;
  alsa_functions.drain			= alsa_drain;

  return(&alsa_functions);
}



static int	alsa_init(const lamipPluginHandle* cfg)
{
  char*	tmp;
  char*	elem_name;

  DEBUG("alsa_init");
  device	= 0;
  handle	= 0;
  if(!cfg)
    return(0);
  /* check configuration values */
  if((tmp=lamip_cfg_get((lamipPluginHandle*)cfg, "device")) == 0){
    lamip_cfg_set((lamipPluginHandle*)cfg, "device", "plughw:0,0");
    tmp = strdup("plughw:0,0");
  }
  device = tmp;

  if((tmp=lamip_cfg_get((lamipPluginHandle*)cfg, "card_id")) == 0){
    lamip_cfg_set((lamipPluginHandle*)cfg, "card_id", "hw:0");
    tmp = strdup("hw:0");
  }
  card_id = tmp;

  if((tmp=lamip_cfg_get((lamipPluginHandle*)cfg, "elem_name")) == 0){
    lamip_cfg_set((lamipPluginHandle*)cfg, "elem_name", "Master");
    tmp = strdup("Master");
  }
  elem_name = tmp;

  /**
   * Is it really necessary since i don't touch anymore to sound volume ?
   *
   */
#if 0
  if((no_mixer=lamip_cfg_get((lamipPluginHandle*)cfg, "no_mixer")) == 0){
    lamip_cfg_set((lamipPluginHandle*)cfg, "no_mixer", "FALSE");
    no_mixer = "FALSE";
  }
  /* check for the card mixer */
  if(!strncasecmp(no_mixer, "False", 5)){
    lamip_send_message("ALSA: mixer setup to FALSE, no mixer init");
    mixer = 0;
    return(1);
  }
  /* setup the apropriate mixer element */
  if((err=snd_mixer_open(&mixer, 0)) < 0){
    lamip_send_message("ALSA: mixer device init failed\n");
    mixer = 0;
    return(0);
  }

  if((err=snd_mixer_attach(mixer, card_id)) < 0){
    lamip_send_message("ALSA; can't attach mixer");
    snd_mixer_close(mixer);
    mixer = 0;
    return(0);
  }
  if((err=snd_mixer_selem_register(mixer, 0, 0)) < 0){
    lamip_send_message("ALSA: Can't register mixer");
    snd_mixer_close(mixer);
    mixer = 0;
    return(0);
  }
  if((err=snd_mixer_load(mixer)) < 0){
    lamip_send_message("ALSA: Can't load mixer");
    snd_mixer_close(mixer);
    mixer = 0;
    return(0);
  }
  snd_mixer_selem_id_malloc(&pcm_elt);
  pcm_element = snd_mixer_first_elem(mixer);
  if(pcm_element == 0){
    lamip_send_message("ALSA: init error no pcm element(s)");
    snd_mixer_selem_id_free(pcm_elt);
    pcm_elt = 0;
    snd_mixer_close(mixer);
    mixer = 0;
    return(0);
  }
  snd_mixer_selem_get_id(pcm_element, pcm_elt);
  while( (elem_name=(char*)snd_mixer_selem_id_get_name(pcm_elt))!=0 ){
    if(!strcmp(elem_name, lamip_cfg_get(cfg, "elem_name"))){
      master_ok = 1;
      break;
    }
    pcm_element = snd_mixer_elem_next(pcm_element);
    if(!pcm_element)
      break;
    snd_mixer_selem_get_id(pcm_element, pcm_elt);
    free(elem_name);
  }
  if(master_ok == 0){
    lamip_send_message("ALSA: no correct sound element entry found");
    snd_mixer_selem_id_free(pcm_elt);
    pcm_elt = 0;
    snd_mixer_close(mixer);
    mixer = 0;
    return(0);
  }
  snd_mixer_selem_set_playback_volume_range(pcm_element, 0, 100);
#endif
  return(1);
}

static void	alsa_cleanup(void)
{
  DEBUG("alsa_cleanup");
  if(device) free(device);
  device = 0;
  if(card_id) free(card_id);
  card_id = 0;
  handle = 0;
}


static int	alsa_open(const lamipPCMInfo* pcmi)
{
  int			err = 0;
  unsigned int		smpl_rate = 0;
  char*			tmp_device = 0;
  int			buffer_time = 0;
  int			period_time = 0;
  snd_output_t*		log;
#ifdef DEBUGING
  snd_pcm_info_t*	info = 0;
#endif

  DEBUG("alsa_open");

  if(pcmi->channels == 1 || pcmi->channels == 2){
    tmp_device = strdup(device);
  }else if(pcmi->channels == 4){
    tmp_device = device;
    device = strdup("surround40");
  }else if(pcmi->channels == 6){
    tmp_device = device;
    device = strdup("surround51");
  }else{
    lamip_send_message("ALSA: no good channel number for output (2/4/6)");
    return(0);
  }
  if((err=snd_pcm_open(&handle, device, SND_PCM_STREAM_PLAYBACK, 0))<0){
    lamip_send_message("ALSA: Playback open error - %s", snd_strerror(err));
    free(device);
    device = tmp_device;
    handle = 0;
    return(0);
  }
  // just for info...
#ifdef DEBUGING
  if(!info) snd_pcm_info_malloc(&info);
  snd_pcm_info(handle, info);
  fprintf(stderr, "ALSA: Card %i, Device %i, Subdevice %i\n",
	  snd_pcm_info_get_card(info),
	  snd_pcm_info_get_device(info),
	  snd_pcm_info_get_subdevice(info));
  snd_pcm_info_free(info);
#endif

  /* For sound logs */
  snd_output_stdio_attach(&log, stderr, 0);

  /* set hardware */
  if((err=snd_pcm_hw_params_malloc(&hwparams)) < 0){
    lamip_send_message("ALSA: cannot allocate hardware parameter structure");
    free(device);
    device = tmp_device;
    return(0);
  }
  if((err=snd_pcm_hw_params_any(handle, hwparams)) < 0){
    lamip_send_message("ALSA: No configuration available");
    free(device);
    device = tmp_device;
    snd_pcm_hw_params_free(hwparams);
    return(0);
  }
  if((err=snd_pcm_hw_params_set_access(handle, hwparams,
				       SND_PCM_ACCESS_RW_INTERLEAVED)) < 0){
    lamip_send_message("ALSA: cannot set access type (%s)",
		       snd_strerror(err));
    free(device);
    device = tmp_device;
    snd_pcm_hw_params_free(hwparams);
    return(0);
  }
#ifdef DEBUGING
  fprintf(stderr, "ALSA: set format (fmt %i)\n", pcmi->format);
#endif
  if(pcmi->format == PCM_FORMAT_S8)
    alsa_format = SND_PCM_FORMAT_S8;
  else if(pcmi->format == PCM_FORMAT_U8)
    alsa_format = SND_PCM_FORMAT_U8;
  else if(pcmi->format == PCM_FORMAT_S16_LE)
    alsa_format = SND_PCM_FORMAT_S16_LE;
  else if(pcmi->format == PCM_FORMAT_S16_BE)
    alsa_format = SND_PCM_FORMAT_S16_BE;
  else if(pcmi->format == PCM_FORMAT_U16_LE)
    alsa_format = SND_PCM_FORMAT_U16_LE;
  else if(pcmi->format == PCM_FORMAT_U16_BE)
    alsa_format = SND_PCM_FORMAT_U16_BE;
  else if(pcmi->format == PCM_FORMAT_S24_LE)
    alsa_format = SND_PCM_FORMAT_S24_3LE;
  else if(pcmi->format == PCM_FORMAT_S24_BE)
    alsa_format = SND_PCM_FORMAT_S24_3BE;
  else if(pcmi->format == PCM_FORMAT_U24_LE)
    alsa_format = SND_PCM_FORMAT_U24_3LE;
  else if(pcmi->format == PCM_FORMAT_U24_BE)
    alsa_format = SND_PCM_FORMAT_U24_3BE;
  else if(pcmi->format == PCM_FORMAT_S32_LE)
    alsa_format = SND_PCM_FORMAT_S32_LE;
  else if(pcmi->format == PCM_FORMAT_S32_BE)
    alsa_format = SND_PCM_FORMAT_S32_BE;
  else if(pcmi->format == PCM_FORMAT_U32_LE)
    alsa_format = SND_PCM_FORMAT_U32_LE;
  else if(pcmi->format == PCM_FORMAT_U32_BE)
    alsa_format = SND_PCM_FORMAT_U32_BE;
  else{
    lamip_send_message("ALSA: setup lamip_pcm_format to alsa_pcm_format error");
    free(device);
    device = tmp_device;
    snd_pcm_hw_params_free(hwparams);
    return(0);
  }
  if((err = snd_pcm_hw_params_set_format(handle, hwparams, alsa_format)) < 0){
    lamip_send_message("ALSA: sample format not available for playback\n");
    free(device);
    device = tmp_device;
    snd_pcm_hw_params_free(hwparams);
    return(0);
  }
  /* set sample rate */
  lamip_send_message("ALSA: set samplerate - %lu", pcmi->samplerate);
  smpl_rate = pcmi->samplerate;
  if((err=snd_pcm_hw_params_set_rate_near(handle, hwparams, &smpl_rate, 0))<0){
    lamip_send_message("ALSA: can't set sample rate");
    free(device);
    device = tmp_device;
    snd_pcm_hw_params_free(hwparams);
    return(0);
  }
  lamip_send_message("ALSA: set channels - %lu", pcmi->channels);
  nb_channel_tmp = pcmi->channels;
  if((err=snd_pcm_hw_params_set_channels(handle, hwparams, pcmi->channels))<0){
    lamip_send_message("ALSA: can't set channels");
    free(device);
    device = tmp_device;
    snd_pcm_hw_params_free(hwparams);
    return(0);
  }
  /* Getting the maximum possible buffer_time */
  if((err = snd_pcm_hw_params_get_buffer_time_max(hwparams, &buffer_time, 0))<0){
    lamip_send_message("ALSA: Unable to get buffer_time_max !");
    free(device);
    device = tmp_device;
    snd_pcm_hw_params_free(hwparams);
    return(0);
  }
  /* Adjusting things... */
  if(buffer_time > 500000)
    buffer_time = 500000;
  if(buffer_time > 0)
    period_time = buffer_time / 4;
  /* Setting period_time */
  if(period_time > 0){
    err = snd_pcm_hw_params_set_period_time_near(handle, hwparams, &period_time, 0);
    if(err < 0){
      lamip_send_message("ALSA: Unable to set period_time near %d!", period_time);
      free(device);
      device = tmp_device;
      snd_pcm_hw_params_free(hwparams);
      return(0);
    }
  }
  /* Setting buffer_time */
  if(buffer_time>0){
    err = snd_pcm_hw_params_set_buffer_time_near(handle, hwparams, &buffer_time, 0);
    if(err<0){
      lamip_send_message("ALSA: Unable to set buffer_time near %d!", buffer_time);
      free(device);
      device = tmp_device;
      snd_pcm_hw_params_free(hwparams);
      return(0);
    }
  } 
  lamip_send_message("ALSA: set format to sound card");
  err = snd_pcm_hw_params(handle, hwparams);
  if(err < 0){
    lamip_send_message("ALSA: Unable to set device");
    free(device);
    device = tmp_device;
    snd_pcm_hw_params_free(hwparams);
    return(0);
  }
  snd_pcm_hw_params_free(hwparams);
  if((err=snd_pcm_prepare(handle))<0){
    lamip_send_message("ALSA: cannot prepare audio interface for use (%s)", snd_strerror(err));
    free(device);
    device = tmp_device;
    return(0);
  }
#ifdef DEBUGING
  snd_pcm_dump(handle, log);
#endif
  free(device);
  device = tmp_device;
  return(1);
}

static void     alsa_close(void)
{
  DEBUG("alsa_close");
  snd_pcm_close(handle);
  handle=0;
}

static size_t	alsa_write_audio(const void* buffer, size_t len)
{
  int			samplesize;
  snd_pcm_sframes_t	written_frames;
  int			num_frames;

  samplesize = snd_pcm_format_physical_width(alsa_format)/8;
  written_frames = 0;
  num_frames = len/(nb_channel_tmp*samplesize);
  while(handle && num_frames>0){
    written_frames = snd_pcm_writei(handle, buffer, num_frames);
    if(written_frames>0){
      num_frames -= written_frames;
      buffer += written_frames*(nb_channel_tmp*samplesize);
    }else if(written_frames == -EPIPE){
      if(handle) snd_pcm_prepare(handle);
    }else{
      if(handle) snd_pcm_prepare(handle);
      break;
    }
  }
  return(len);
}

static void	alsa_drain(void)
{
  if(handle) snd_pcm_drain(handle);
}

static void	alsa_drop(void)
{
  if(handle) snd_pcm_drop(handle);
}
